/**
 * \file imagebox.cpp
 * A widget that encapsulates operations with cover art pictures
 *
 * \b Project: Qoobar
 * \author Alex Novichkov
 * \date 11 Aug 2011
 *
 * Copyright (C) 2011  Alex Novichkov
 *
 * This file is part of Qoobar.
 *
 * Qoobar is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Qoobar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "imagebox.h"
#include "imagedialog.h"
#include "application.h"

#ifdef HAVE_QT5
#include <QtWidgets>
#else
#include <QtGui>
#endif

#include "enums.h"

Panel::Panel(QWidget *parent) : QWidget(parent)
{
    label = new ClickableLabel(this);
    connect(label,SIGNAL(clicked()),SLOT(showFullImage()));
    label->setContentsMargins(5,5,0,5);
    panel = new FadingPanel(this);
    QHBoxLayout *layout = new QHBoxLayout;

    layout->setContentsMargins(0,0,0,0);
    layout->addWidget(label);
    layout->addStretch();
    layout->addWidget(panel);
    setLayout(layout);
    empty=true;
}

void Panel::showFullImage()
{
    if (empty || image.isNull()) return;

    ImageDialog *dialog = new ImageDialog(this);
    dialog->setImage(image,toolTip());
    dialog->exec();
}

void Panel::setText(QString text, bool clip/*=false*/)
{
    empty = false;
    image = QPixmap();
    if (clip) text=this->fontMetrics().elidedText(text,Qt::ElideRight,150);
    label->setText(text);
}

void Panel::setPixmap(QPixmap p, const QString &toolTip)
{
    setToolTip(toolTip);
    label->setText(QString());
    empty=false;
    image=p;
    int max=qMax(p.width(),p.height());
    if (max>150) {
        if (p.width()==max) p=p.scaledToWidth(150);
        else p=p.scaledToHeight(150);
    }
    label->setPixmap(p);
}

void Panel::clear()
{
    empty=true;
    label->clear();
}

void Panel::enterEvent(QEvent * event)
{
    QWidget::enterEvent(event);
#ifndef Q_OS_MAC
    changeHoverState(true);
#endif
}

void Panel::leaveEvent(QEvent * event)
{
    QWidget::leaveEvent(event);
#ifndef Q_OS_MAC
    changeHoverState(false);
#endif
}

void Panel::paintEvent(QPaintEvent *paintEvent)
{
    QWidget::paintEvent(paintEvent);
#ifndef Q_OS_MAC
    QPainter p(this);

    QPoint topLeft(label->geometry().left(), contentsRect().top());
    const QRect paintArea(topLeft, contentsRect().bottomRight());

    if (backgroundPixmap.isNull() || backgroundPixmap.size() != size())
        backgroundPixmap = cacheBackground(paintArea.size());
    p.drawPixmap(paintArea, backgroundPixmap);
#endif
}

QPixmap Panel::cacheBackground(const QSize &size)
{
    QPixmap pixmap(size);
    pixmap.fill(Qt::transparent);
    QPainter p(&pixmap);

    QRect fullRect(0, 0, size.width(), size.height());
    p.fillRect(fullRect, QColor(255, 255, 255, 40));

    QLinearGradient lg(fullRect.topLeft(), fullRect.bottomLeft());
    lg.setColorAt(0, QColor(255, 255, 255, 130));
    lg.setColorAt(1, QColor(255, 255, 255, 0));
    p.fillRect(fullRect, lg);
    p.setRenderHint(QPainter::Antialiasing, true);
    p.translate(0.5, 0.5);
    p.setPen(QColor(0, 0, 0, 40));
    p.setBrush(Qt::NoBrush);
    p.drawRoundedRect(fullRect.adjusted(0, 0, -1, -1), 2, 2);
    p.setPen(QColor(255,255,255,140));
    p.drawRoundedRect(fullRect.adjusted(1, 1, -2, -2), 2, 2);

    return pixmap;
}

void Panel::changeHoverState(bool hovered)
{
    panel->fadeTo(hovered ? 1.0 : 0);
}

void Panel::addWidget(QWidget *w)
{
    panel->layout()->addWidget(w);
}

FadingPanel::FadingPanel(QWidget *parent) :
    QWidget(parent)
{
#ifndef Q_OS_MAC
#if QT_VERSION >= 0x040600
    m_opacityEffect=new QGraphicsOpacityEffect;
    m_opacityEffect->setOpacity(0);
    setGraphicsEffect(m_opacityEffect);
#endif
    QPalette pal;
    pal.setBrush(QPalette::All, QPalette::Window, Qt::transparent);
    setPalette(pal);
#endif
    QVBoxLayout *l = new QVBoxLayout;
    l->setContentsMargins(0,0,0,0);
    l->addStretch();
    setLayout(l);
}

void FadingPanel::fadeTo(float value)
{
#if QT_VERSION >= 0x040600
    QPropertyAnimation *animation = new QPropertyAnimation(m_opacityEffect, "opacity");
    animation->setDuration(200);
    animation->setEndValue(value);
    animation->start(QAbstractAnimation::DeleteWhenStopped);
#endif
}

ImageBox::ImageBox(QWidget *parent) : QWidget(parent)
{
    addImageAct = new QAction(this);
    removeImageAct = new QAction(this);
    saveImageAct = new QAction(this);
    QToolButton *addImageButton = makeAction(addImageAct,
                                             SLOT(addImage()),QSL("list-add"),
                                             QSL(":/src/icons/list-add1.png"));
    QToolButton *removeImageButton = makeAction(removeImageAct,
                                                SLOT(removeImage()),QSL("list-remove"),
                                                QSL(":/src/icons/list-remove.png"));
    QToolButton *saveImageButton = makeAction(saveImageAct,
                                              SLOT(saveImage()),QSL("document-save"),
                                              QSL(":/src/icons/document-save.png"));
    addImageAct->setEnabled(true);
    removeImageAct->setEnabled(true);
    saveImageAct->setEnabled(true);

    //type
    typePanel = new Panel(this);
    typeAct = new QAction(this);
    typePanel->addWidget(makeAction(typeAct,SLOT(changeType()),QString(),
                                    QSL(":/src/icons/TextEdit.png")));

    //description
    descriptionPanel = new Panel(this);
    descriptionAct = new QAction(this);
    descriptionPanel->addWidget(makeAction(descriptionAct,
                                           SLOT(changeDescription()),QString(),
                                           QSL(":/src/icons/TextEdit.png")));

    imagePanel = new Panel(this);
    cutAct = new QAction(this);
    copyAct = new QAction(this);
    pasteAct = new QAction(this);
    imagePanel->addWidget(makeAction(cutAct,  SLOT(cutImage()),
                                     QSL("edit-cut"), QSL(":/src/icons/edit-cut.png")));
    imagePanel->addWidget(makeAction(copyAct, SLOT(copyImage()),
                                     QSL("edit-copy"), QSL(":/src/icons/edit-copy.png")));
    imagePanel->addWidget(makeAction(pasteAct, SLOT(pasteImage()),
                                     QSL("edit-paste"), QSL(":/src/icons/edit-paste.png")));

    QGridLayout *imageLayout=new QGridLayout;
#ifdef Q_OS_MAC
    imageLayout->setContentsMargins(0,3,3,3);
#else
    imageLayout->setContentsMargins(0,0,0,0);
#endif
    imageLayout->setVerticalSpacing(0);
    imageLayout->addWidget(typePanel,0,0,1,3);
    imageLayout->addWidget(descriptionPanel,1,0,1,3);
    imageLayout->addWidget(imagePanel,2,0,1,3,Qt::AlignTop);
    imageLayout->addWidget(addImageButton,3,0,1,1,Qt::AlignCenter);
    imageLayout->addWidget(removeImageButton,3,1,1,1,Qt::AlignCenter);
    imageLayout->addWidget(saveImageButton,3,2,1,1,Qt::AlignCenter);

    imageLayout->setRowStretch(0,1);
    imageLayout->setRowStretch(1,1);
    imageLayout->setRowStretch(2,20);
    imageLayout->setRowStretch(3,1);
    setLayout(imageLayout);
}

void ImageBox::retranslateUi()
{
    addImageAct->setText(tr("Add an image"));
    removeImageAct->setText(tr("Remove the image"));
    saveImageAct->setText(tr("Export the image"));

    typeAct->setText(tr("Edit type"));
    descriptionAct->setText(tr("Edit description"));

    cutAct->setText(tr("Cut"));
    copyAct->setText(tr("Copy"));
    pasteAct->setText(tr("Paste"));

    mp3CoverTypes.clear();
    mp3CoverTypes << tr("Other");
    mp3CoverTypes << tr("File Icon");
    mp3CoverTypes << tr("Other File Icon");
    mp3CoverTypes << tr("Front Cover");
    mp3CoverTypes << tr("Back Cover");
    mp3CoverTypes << tr("Leaflet Page");
    mp3CoverTypes << tr("Media");
    mp3CoverTypes << tr("Lead Artist");
    mp3CoverTypes << tr("Artist");
    mp3CoverTypes << tr("Conductor");
    mp3CoverTypes << tr("Band");
    mp3CoverTypes << tr("Composer");
    mp3CoverTypes << tr("Lyricist");
    mp3CoverTypes << tr("Recording Location");
    mp3CoverTypes << tr("During Recording");
    mp3CoverTypes << tr("During Performance");
    mp3CoverTypes << tr("Movie Screen Capture");
    mp3CoverTypes << tr("Coloured Fish");
    mp3CoverTypes << tr("Illustration");
    mp3CoverTypes << tr("Band Logo");
    mp3CoverTypes << tr("Publisher Logo");

    updatePicture(image,filePath);
}

void ImageBox::changeType()
{
    if (image.pixmap().isEmpty()) return;
    int type = image.type();
    QString item = QInputDialog::getItem(this,tr("Qoobar"),
                                         tr("Cover art type"),
                                         mp3CoverTypes,type,false);
    if (item.isEmpty()) return;
    type = mp3CoverTypes.indexOf(item);

    if (image.type()!=type) {
        CoverImage img=image;
        img.setType(type);
        emit imageChanged(img,tr("changing cover art"));
    }
}

void ImageBox::changeDescription()
{
    if (image.pixmap().isEmpty()) return;
    QString description = image.description();

    bool ok;
    QString item = QInputDialog::getText(this,tr("Qoobar"),
                                         tr("Cover art description"),QLineEdit::Normal,
                                         description,&ok);
    if (description!=item && ok) {
        CoverImage img=image;
        img.setDescription(item);
        emit imageChanged(img,tr("changing cover art"));
    }
}

void ImageBox::addImage()
{

    //get the folder by the first selected file
    QString filename=QFileDialog::getOpenFileName(this,tr("Adding cover art"),filePath,
                                                  tr("Image files (*.png *.jpg *.jpeg)"));
    if (filename.isEmpty()) return;


    QFile f(filename);
    QFileInfo fi(filename);
    f.open(QFile::ReadOnly);
    QByteArray newPix=f.readAll();
    if (!newPix.isEmpty()) {
        CoverImage img(newPix,3,QSL("image/png"),fi.fileName());
        if (filename.endsWith(QLS("jpg"),Qt::CaseInsensitive)
                || filename.endsWith(QLS("jpeg"),Qt::CaseInsensitive))
            img.setMimetype(QSL("image/jpeg"));

        emit imageChanged(img,tr("adding cover art"));
    }
}

void ImageBox::removeImage()
{
    CoverImage img;
    emit imageChanged(img,tr("removing cover art"));
}

QToolButton *ImageBox::makeAction(QAction *act, const char *slot,
                       const QString &themeIcon, const QString &icon)
{
    connect(act,SIGNAL(triggered()),slot);
    QToolButton *button = new QToolButton(this);
    button->setAutoRaise(true);
    button->setDefaultAction(act);
#if QT_VERSION >= 0x040600
    act->setIcon(QIcon::fromTheme(themeIcon,QIcon(icon)));
#else
    act->setIcon(QIcon(icon));
#endif
    act->setEnabled(false);
    return button;
}

void ImageBox::saveImage()
{
    if (!image.pixmap().isNull()) {
        QString filter;
        QString filename=QFileDialog::getSaveFileName(this,
                         tr("Exporting cover art"),
                         QFileInfo(filePath).path(),
                         tr("JPEG files (*.jpg *.jpeg);;PNG files (*.png);;BMP files (*.bmp);;XPM files (*.xpm)"),
                         &filter);
        if (!filename.isEmpty()) {
            if (filter.isEmpty())
                filename.append(".png");
            QImage img;
            img.loadFromData(image.pixmap());
            img.save(filename);
        }
    }
}

void ImageBox::copyImage()
{
    App->imageBuffer = image;
    pasteAct->setEnabled(true);
}

void ImageBox::cutImage()
{
    copyImage();
    removeImage();
}

void ImageBox::pasteImage()
{
    emit imageChanged(App->imageBuffer,tr("changing cover art"));
}

void ImageBox::clear()
{
    imagePanel->clear();
    image.clear();
    filePath.clear();
    typePanel->clear();
    descriptionPanel->clear();
    cutAct->setEnabled(false);
    copyAct->setEnabled(false);
    pasteAct->setEnabled(App->imageBuffer != CoverImage());
    descriptionAct->setEnabled(false);
    typeAct->setEnabled(false);
}

void ImageBox::updatePicture(const QString &text)
{
    imagePanel->setText(text);
}

void ImageBox::updatePicture(const CoverImage &img, const QString &filename)
{
    image=img;
    filePath=filename;
    if (!img.pixmap().isNull()) {
        QPixmap x;
        x.loadFromData(img.pixmap());

        QStringList l=img.mimetype().split(QSL("/"));
        QString tooltip=QString("%1x%2").arg(x.width()).arg(x.height());
        if (l.size()>1)
            tooltip.prepend(QString("%1, ").arg(l.at(1).toUpper()));

        typePanel->setText(mp3CoverTypes.value(img.type()));
        descriptionPanel->setText(img.description(),true);
        imagePanel->setPixmap(x,tooltip);
        cutAct->setEnabled(true);
        copyAct->setEnabled(true);
        pasteAct->setEnabled(App->imageBuffer != CoverImage());
        descriptionAct->setEnabled(true);
        typeAct->setEnabled(true);
    }
}

